
object AttributeClosure{

  /*	
   *  fdependencies		: all the functional dependencies in a relation
   * 											MUST BE in irreducible form (at least left/right)
   *                      
   *  closureAttrs    : set over which we will compute the closure
   *  
   *  RETURN
   *  	 a two-tuple having the closure set over input set closureAttrs and 
   *       a boolean telling whether the input set is a superkey of the relation 
  */
  def attributeClosure( fdependencies   :  Set[Tuple2[Set[Char],Char]],  
                        closureAttrs    :  Set[Char] ): ( Set[Char], Boolean ) = {  
    
    // ensure the algorithm is tail recursive so the compiler can translate it 
    // into byte code using a single stack frame
    import annotation.tailrec
    @tailrec def go( attrClosureSet: Set[Char] ) : Set[Char] = {      

      // for all functional dependencies in the relation, get the (flattened) sets 
      // of dependents for the current set of determinants
      val thisRoundClosure: Set[Char] = 
                                 fdependencies
                                    .withFilter( _._1.subsetOf( attrClosureSet ) )
                                    .flatMap( fdep => Set(fdep._2) )        

      if( thisRoundClosure.diff( attrClosureSet ).isEmpty)
        // no new attributes were added; return last set built
        { attrClosureSet }
      else
        // add the newly reached dependents to the closure set being built 
        // and recurse 
        { go( attrClosureSet.union( thisRoundClosure ) ) }
    }
        
    val closure: Set[Char] = go( closureAttrs )  // invoke the nested computation
    
    // from all attributes, dependent or determinant or both, compute the set
    // of all relation attributes; used next to determine if closure is superkey
    val allAttributes: Set[Char] = fdependencies.flatMap( _._1 ).
                                                        union( fdependencies.map(_._2 ) )
    
    // return the closure starting from closureAttrs and whether or not it is a 
    // superkey of the relation
    (closure, 
      ( if( allAttributes == (closure intersect allAttributes) ) true else false ) )
  }
}

/*---------------------*/

object AttrC_test extends App{
  
  // internal representation of dependency diagram
  val fdep: Set[Tuple2[Set[Char],Char]] = Set( 
                                            (Set('R'), 'T'),
                                            (Set('R'), 'D'),
                                            (Set('R'), 'B'),
                                            (Set('L'), 'Y'),
                                            (Set('L'), 'A'),               
                                            (Set('C','S'), 'G'),
                                            (Set('Y'), 'C'),
                                            (Set('S'), 'O'),
                                            (Set('L'), 'R'),
                                            (Set('E','Z'), 'L'),
                                            (Set('Y'), 'S'),
                                            (Set('L'), 'Z'),                                            
                                            (Set('H'), 'L'),
                                            (Set('S'), 'N'),
                                            (Set('L'), 'H'),
                                            (Set('S','C'), 'P'),
                                            (Set('C','S'), 'Y'),
                                            (Set('L'), 'E')
                                            )
  
  import AttributeClosure._
  
  /* test series of attribute closures */
  
  println
  println("/*--------------------------*/")

  /*-------TEST A SERIES OF IMPORTANT CASES-------*/
    
  val closureAttrs_R: Set[Char] = Set('R')
  println( s"closure over {R}:     ${attributeClosure( fdep,closureAttrs_R )}" )

  println("/*--------------------------*/")  
  
  val closureAttrs_Y: Set[Char] = Set('Y')
  println( s"closure over {Y}:     ${attributeClosure( fdep,closureAttrs_Y )}" )

  println("/*--------------------------*/")
  
  val closureAttrs_CS: Set[Char] = Set('C','S')  
  println( s"closure over {CS}:    ${attributeClosure( fdep,closureAttrs_CS )}" )
  
  println("/*--------------------------*/")
  
  val closureAttrs_RSC: Set[Char] = Set('R','S','C')
  println( s"closure over {RSC}:   ${attributeClosure( fdep,closureAttrs_RSC )}" )
  
  println("/*--------------------------*/")
  
  val closureAttrs_L: Set[Char] = Set('L')
  println( s"closure over {L}:     ${attributeClosure( fdep,closureAttrs_L )}" )
  
  println("/*--------------------------*/")
  
  val closureAttrs_EZ: Set[Char] = Set('E','Z')
  println( s"closure over {EZ}:    ${attributeClosure( fdep,closureAttrs_EZ )}" )
  
  println("/*--------------------------*/")
  
  val closureAttrs_RSCL: Set[Char] = Set('R','L','C','S')
  println( s"closure over {RSCL}:  ${attributeClosure( fdep,closureAttrs_RSCL )}" )
  
  println("/*--------------------------*/")
  
  val closureAttrs_RSCAL: Set[Char] = Set('R','L','A','C','S')
  println( s"closure over {RSCAL}: ${attributeClosure( fdep,closureAttrs_RSCAL )}" )
}